@@ -43,14 +43,19 @@
 
 /* 0x100-0x11f reserved for spice, 0x1ff used for unstable work */
 #define QXL_DEVICE_ID_STABLE 0x0100
-#define QXL_REVISION_STABLE_V04 0x01
-#define QXL_REVISION_STABLE_V06 0x02
+
+enum {
+    QXL_REVISION_STABLE_V04=0x01,
+    QXL_REVISION_STABLE_V06=0x02,
+    QXL_REVISION_STABLE_V10=0x03,
+    QXL_REVISION_STABLE_V12=0x04,
+};
 
 #define QXL_DEVICE_ID_DEVEL 0x01ff
 #define QXL_REVISION_DEVEL 0x01
 
-#define QXL_ROM_MAGIC (*(uint32_t*)"QXRO")
-#define QXL_RAM_MAGIC (*(uint32_t*)"QXRA")
+#define QXL_ROM_MAGIC (*(const uint32_t*)"QXRO")
+#define QXL_RAM_MAGIC (*(const uint32_t*)"QXRA")
 
 enum {
     QXL_RAM_RANGE_INDEX,
@@ -80,6 +85,17 @@ enum {
     QXL_IO_DESTROY_PRIMARY,
     QXL_IO_DESTROY_SURFACE_WAIT,
     QXL_IO_DESTROY_ALL_SURFACES,
+    /* appended for qxl-3 */
+    QXL_IO_UPDATE_AREA_ASYNC,
+    QXL_IO_MEMSLOT_ADD_ASYNC,
+    QXL_IO_CREATE_PRIMARY_ASYNC,
+    QXL_IO_DESTROY_PRIMARY_ASYNC,
+    QXL_IO_DESTROY_SURFACE_ASYNC,
+    QXL_IO_DESTROY_ALL_SURFACES_ASYNC,
+    QXL_IO_FLUSH_SURFACES_ASYNC,
+    QXL_IO_FLUSH_RELEASE,
+    /* appended for qxl-4 */
+    QXL_IO_MONITORS_CONFIG_ASYNC,
 
     QXL_IO_RANGE_SIZE
 };
@@ -109,6 +125,13 @@ typedef struct SPICE_ATTR_PACKED QXLRect {
     int32_t right;
 } QXLRect;
 
+typedef struct SPICE_ATTR_PACKED QXLURect {
+    uint32_t top;
+    uint32_t left;
+    uint32_t bottom;
+    uint32_t right;
+} QXLURect;
+
 /* qxl-1 compat: append only */
 typedef struct SPICE_ATTR_PACKED QXLRom {
     uint32_t magic;
@@ -132,9 +155,19 @@ typedef struct SPICE_ATTR_PACKED QXLRom {
     uint8_t slot_gen_bits;
     uint8_t slot_id_bits;
     uint8_t slot_generation;
-    uint8_t padding[3]; /* Padding to 32bit align */
+    /* appended for qxl-4 */
+    uint8_t client_present;
+    uint8_t client_capabilities[58];
+    uint32_t client_monitors_config_crc;
+    struct {
+        uint16_t count;
+        uint16_t padding;
+        QXLURect heads[64];
+    } client_monitors_config;
 } QXLRom;
 
+#define CLIENT_MONITORS_CONFIG_CRC32_POLY 0xedb88320
+
 /* qxl-1 compat: fixed */
 typedef struct SPICE_ATTR_PACKED QXLMode {
     uint32_t id;
@@ -200,15 +233,23 @@ typedef struct SPICE_ATTR_PACKED QXLSurfaceCreate {
     QXLPHYSICAL mem;
 } QXLSurfaceCreate;
 
-SPICE_RING_DECLARE(QXLCommandRing, QXLCommand, 32);
-SPICE_RING_DECLARE(QXLCursorRing, QXLCommand, 32);
+#define QXL_COMMAND_RING_SIZE 32
+#define QXL_CURSOR_RING_SIZE 32
+#define QXL_RELEASE_RING_SIZE 8
 
-SPICE_RING_DECLARE(QXLReleaseRing, uint64_t, 8);
+SPICE_RING_DECLARE(QXLCommandRing, QXLCommand, QXL_COMMAND_RING_SIZE);
+SPICE_RING_DECLARE(QXLCursorRing, QXLCommand, QXL_CURSOR_RING_SIZE);
+
+SPICE_RING_DECLARE(QXLReleaseRing, uint64_t, QXL_RELEASE_RING_SIZE);
 
 #define QXL_LOG_BUF_SIZE 4096
 
 #define QXL_INTERRUPT_DISPLAY (1 << 0)
 #define QXL_INTERRUPT_CURSOR (1 << 1)
+#define QXL_INTERRUPT_IO_CMD (1 << 2)
+#define QXL_INTERRUPT_ERROR  (1 << 3)
+#define QXL_INTERRUPT_CLIENT (1 << 4)
+#define QXL_INTERRUPT_CLIENT_MONITORS_CONFIG  (1 << 5)
 
 /* qxl-1 compat: append only */
 typedef struct SPICE_ATTR_PACKED QXLRam {
@@ -225,6 +266,12 @@ typedef struct SPICE_ATTR_PACKED QXLRam {
     QXLMemSlot mem_slot;
     QXLSurfaceCreate create_surface;
     uint64_t flags;
+
+    /* appended for qxl-4 */
+
+    /* used by QXL_IO_MONITORS_CONFIG_ASYNC */
+    QXLPHYSICAL monitors_config;
+
 } QXLRam;
 
 typedef union QXLReleaseInfo {
@@ -319,6 +366,7 @@ enum {
     QXL_DRAW_TEXT,
     QXL_DRAW_TRANSPARENT,
     QXL_DRAW_ALPHA_BLEND,
+    QXL_DRAW_COMPOSITE
 };
 
 typedef struct SPICE_ATTR_PACKED QXLRasterGlyph {
@@ -459,6 +507,79 @@ typedef struct SPICE_ATTR_PACKED QXLClip {
     QXLPHYSICAL data;
 } QXLClip;
 
+typedef enum {
+    QXL_OP_CLEAR                     = 0x00,
+    QXL_OP_SOURCE		     = 0x01,
+    QXL_OP_DST                       = 0x02,
+    QXL_OP_OVER                      = 0x03,
+    QXL_OP_OVER_REVERSE              = 0x04,
+    QXL_OP_IN                        = 0x05,
+    QXL_OP_IN_REVERSE                = 0x06,
+    QXL_OP_OUT                       = 0x07,
+    QXL_OP_OUT_REVERSE               = 0x08,
+    QXL_OP_ATOP                      = 0x09,
+    QXL_OP_ATOP_REVERSE              = 0x0a,
+    QXL_OP_XOR                       = 0x0b,
+    QXL_OP_ADD                       = 0x0c,
+    QXL_OP_SATURATE                  = 0x0d,
+    /* Note the jump here from 0x0d to 0x30 */
+    QXL_OP_MULTIPLY                  = 0x30,
+    QXL_OP_SCREEN                    = 0x31,
+    QXL_OP_OVERLAY                   = 0x32,
+    QXL_OP_DARKEN                    = 0x33,
+    QXL_OP_LIGHTEN                   = 0x34,
+    QXL_OP_COLOR_DODGE               = 0x35,
+    QXL_OP_COLOR_BURN                = 0x36,
+    QXL_OP_HARD_LIGHT                = 0x37,
+    QXL_OP_SOFT_LIGHT                = 0x38,
+    QXL_OP_DIFFERENCE                = 0x39,
+    QXL_OP_EXCLUSION                 = 0x3a,
+    QXL_OP_HSL_HUE                   = 0x3b,
+    QXL_OP_HSL_SATURATION            = 0x3c,
+    QXL_OP_HSL_COLOR                 = 0x3d,
+    QXL_OP_HSL_LUMINOSITY            = 0x3e
+} QXLOperator;
+
+typedef struct {
+    uint32_t	t00;
+    uint32_t	t01;
+    uint32_t	t02;
+    uint32_t	t10;
+    uint32_t	t11;
+    uint32_t	t12;
+} QXLTransform;
+
+/* The flags field has the following bit fields:
+ *
+ *     operator:		[  0 -  7 ]
+ *     src_filter:		[  8 - 10 ]
+ *     mask_filter:		[ 11 - 13 ]
+ *     src_repeat:		[ 14 - 15 ]
+ *     mask_repeat:		[ 16 - 17 ]
+ *     component_alpha:		[ 18 - 18 ]
+ *     reserved:		[ 19 - 31 ]
+ *
+ * The repeat and filter values are those of pixman:
+ *		REPEAT_NONE =		0
+ *              REPEAT_NORMAL =		1
+ *		REPEAT_PAD =		2
+ *		REPEAT_REFLECT =	3
+ *
+ * The filter values are:
+ *		FILTER_NEAREST =	0
+ *		FILTER_BILINEAR	=	1
+ */
+typedef struct SPICE_ATTR_PACKED QXLComposite {
+    uint32_t		flags;
+
+    QXLPHYSICAL		src;
+    QXLPHYSICAL		src_transform;		/* May be NULL */
+    QXLPHYSICAL		mask;			/* May be NULL */
+    QXLPHYSICAL		mask_transform;		/* May be NULL */
+    QXLPoint16		src_origin;
+    QXLPoint16		mask_origin;
+} QXLComposite;
+
 typedef struct SPICE_ATTR_PACKED QXLCompatDrawable {
     QXLReleaseInfo release_info;
     uint8_t effect;
@@ -511,6 +632,7 @@ typedef struct SPICE_ATTR_PACKED QXLDrawable {
         QXLBlackness blackness;
         QXLInvers invers;
         QXLWhiteness whiteness;
+	QXLComposite composite;
     } u;
 } QXLDrawable;
 
@@ -634,6 +756,26 @@ typedef struct SPICE_ATTR_PACKED QXLImage {
     };
 } QXLImage;
 
+/* A QXLHead is a single monitor output backed by a QXLSurface.
+ * x and y offsets are unsigned since they are used in relation to
+ * the given surface, not the same as the x, y coordinates in the guest
+ * screen reference frame. */
+typedef struct SPICE_ATTR_PACKED QXLHead {
+    uint32_t id;
+    uint32_t surface_id;
+    uint32_t width;
+    uint32_t height;
+    uint32_t x;
+    uint32_t y;
+    uint32_t flags;
+} QXLHead;
+
+typedef struct SPICE_ATTR_PACKED QXLMonitorsConfig {
+    uint16_t count;
+    uint16_t max_allowed; /* If it is 0 no fixed limit is given by the driver */
+    QXLHead heads[0];
+} QXLMonitorsConfig;
+
 #include <spice/end-packed.h>
 
 #endif /* _H_QXL_DEV */
